#include <freertos/FreeRTOS.h>
#include <esp_log.h>
#include "imumpu6886.h"
#include "driver/i2c.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "stdint.h"
#include <math.h>

// Definitions
#define RAD_TO_DEG 57.295779513082320876798154814105
#define sampleFreq	500.0f		// sample frequency in Hz
#define betaDef		10.0f		// 2 * proportional gain


// Variable definitions

volatile float beta = betaDef;						// 2 * proportional gain (Kp)
volatile static float q0 = 1.0f, q1 = 0.0f, q2 = 0.0f, q3 = 0.0f;	// quaternion of sensor frame relative to auxiliary frame
/*NVS handler for offset data reading*/
nvs_handle_t my_handle; 

//---------------------------------------------------------------------------------------------------
// Function declarations

static float invSqrt(float x);

//====================================================================================================
// Functions


void MadgwickAHRSetBeta(float beta_in) {
	beta = beta_in;
}

// IMU algorithm update
void MadgwickAHRSupdateIMU(float gx, float gy, float gz, float ax, float ay, float az, float *pitch,float *roll,float *yaw) {
	float recipNorm;
	float s0, s1, s2, s3;
	float qDot1, qDot2, qDot3, qDot4;
	float _2q0, _2q1, _2q2, _2q3, _4q0, _4q1, _4q2 ,_8q1, _8q2, q0q0, q1q1, q2q2, q3q3;

	// Rate of change of quaternion from gyroscope
	qDot1 = 0.5f * (-q1 * gx - q2 * gy - q3 * gz);
	qDot2 = 0.5f * (q0 * gx + q2 * gz - q3 * gy);
	qDot3 = 0.5f * (q0 * gy - q1 * gz + q3 * gx);
	qDot4 = 0.5f * (q0 * gz + q1 * gy - q2 * gx);

	// Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
	if(!((ax == 0.0f) && (ay == 0.0f) && (az == 0.0f))) {

		// Normalise accelerometer measurement
		recipNorm = invSqrt(ax * ax + ay * ay + az * az);
		ax *= recipNorm;
		ay *= recipNorm;
		az *= recipNorm;   

		// Auxiliary variables to avoid repeated arithmetic
		_2q0 = 2.0f * q0;
		_2q1 = 2.0f * q1;
		_2q2 = 2.0f * q2;
		_2q3 = 2.0f * q3;
		_4q0 = 4.0f * q0;
		_4q1 = 4.0f * q1;
		_4q2 = 4.0f * q2;
		_8q1 = 8.0f * q1;
		_8q2 = 8.0f * q2;
		q0q0 = q0 * q0;
		q1q1 = q1 * q1;
		q2q2 = q2 * q2;
		q3q3 = q3 * q3;

		// Gradient decent algorithm corrective step
		s0 = _4q0 * q2q2 + _2q2 * ax + _4q0 * q1q1 - _2q1 * ay;
		s1 = _4q1 * q3q3 - _2q3 * ax + 4.0f * q0q0 * q1 - _2q0 * ay - _4q1 + _8q1 * q1q1 + _8q1 * q2q2 + _4q1 * az;
		s2 = 4.0f * q0q0 * q2 + _2q0 * ax + _4q2 * q3q3 - _2q3 * ay - _4q2 + _8q2 * q1q1 + _8q2 * q2q2 + _4q2 * az;
		s3 = 4.0f * q1q1 * q3 - _2q1 * ax + 4.0f * q2q2 * q3 - _2q2 * ay;
		recipNorm = invSqrt(s0 * s0 + s1 * s1 + s2 * s2 + s3 * s3); // normalise step magnitude
		s0 *= recipNorm;
		s1 *= recipNorm;
		s2 *= recipNorm;
		s3 *= recipNorm;

    //Serial.print(" beta: ");
    //Serial.println(beta);
		// Apply feedback step
		qDot1 -= beta * s0;
		qDot2 -= beta * s1;
		qDot3 -= beta * s2;
		qDot4 -= beta * s3;

	}

	// Integrate rate of change of quaternion to yield quaternion
	q0 += qDot1 * (1.0f / sampleFreq);
	q1 += qDot2 * (1.0f / sampleFreq);
	q2 += qDot3 * (1.0f / sampleFreq);
	q3 += qDot4 * (1.0f / sampleFreq);

	// Normalise quaternion
	recipNorm = invSqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3);
	q0 *= recipNorm;
	q1 *= recipNorm;
	q2 *= recipNorm;
	q3 *= recipNorm;

	*pitch = asin(-2 * q1 * q3 + 2 * q0* q2);	// pitch
	*roll  = atan2(2 * q2 * q3 + 2 * q0 * q1, -2 * q1 * q1 - 2 * q2* q2 + 1);	// roll
	*yaw   = atan2(2*(q1*q2 + q0*q3),q0*q0+q1*q1-q2*q2-q3*q3);	//yaw

	*pitch *= RAD_TO_DEG;
    *yaw   *= RAD_TO_DEG;
    // Declination of SparkFun Electronics (40°05'26.6"N 105°11'05.9"W) is
    // 	8° 30' E  ± 0° 21' (or 8.5°) on 2016-07-19
    // - http://www.ngdc.noaa.gov/geomag-web/#declination
    *yaw   -= 8.5;
    *roll  *= RAD_TO_DEG;
}

//---------------------------------------------------------------------------------------------------
// Fast inverse square-root
// See: http://en.wikipedia.org/wiki/Fast_inverse_square_root

static float invSqrt(float x) {
	float halfx = 0.5f * x;
	float y = x;
	long i = *(long*)&y;
	i = 0x5f3759df - (i>>1);
	y = *(float*)&i;
	y = y * (1.5f - (halfx * y * y));
	return y;
}

static int imuinitf = 0;
static volatile float angle;
SemaphoreHandle_t angle_lock = NULL;
//static void ImuUpdateTask(void *arg);
SemaphoreHandle_t i2c_mutex;
int16_t gryo_x_offset;
int16_t gryo_y_offset;
int16_t gryo_z_offset;

static float acc_x_2, acc_y_2, acc_z_2;
// maybe not need
typedef struct {
    float a1;
    float a2;

    float b0;
    float b1;
    float b2;

    float y1;
    float y2;

    float x1;
    float x2;

} Butter_t;

float ButterFilter2(Butter_t *butter, float input) {
    float output;
    output = butter->b0 * input - butter->a2 * butter->y2 +
             butter->b1 * butter->x1 + butter->b2 * butter->x2 -
             butter->a1 * butter->y1;
    butter->y2 = butter->y1;
    butter->y1 = output;
    butter->x2 = butter->x1;
    butter->x1 = input;
    return output;
}
void SetMutex(SemaphoreHandle_t* mutex) {
    i2c_mutex = *mutex;
}
void getAngle(real32_T* balaange) {

    if(imuinitf==0){
       Imuinit();
	vTaskDelay(pdMS_TO_TICKS(500));
       imuinitf = 1;
    }
    float angle_out;
    xSemaphoreTake(angle_lock, portMAX_DELAY);
    angle_out = angle;
    xSemaphoreGive(angle_lock);
    //return angle_out;
    *balaange = angle_out;
    // printf("angle_out:%f\n", angle_out);
}

/*NVS data reading functions*/
bool preferences_begin(const char * name, bool readOnly, const char* partition_label)
{
  
  esp_err_t err = ESP_OK;
  err = nvs_flash_init();

  if (partition_label != NULL) {
      err = nvs_flash_init_partition(partition_label);
      if (err) {
          ESP_LOGE("ESP32_NVS","nvs_flash_init_partition failed: %s", esp_err_to_name(err));
          return false;
      }
      err = nvs_open_from_partition(partition_label, name, readOnly ? NVS_READONLY : NVS_READWRITE, &my_handle);
  } else {
      err = nvs_open(name, readOnly ? NVS_READONLY : NVS_READWRITE, &my_handle);
  }
  if(err){
      ESP_LOGE("ESP32_NVS","nvs_open failed: %s", esp_err_to_name(err));
      return false;
  }
  
  return true;
} 
void calibrationInit() {
    preferences_begin("Bala2Cal", false, NULL);
}
int32_t preferences_getInt(const char* key, const int32_t defaultValue)
{
   int32_t value = defaultValue;
    if(!key){
        return value;
    }
    esp_err_t err = nvs_get_i32(my_handle, key, &value);
    if(err){
        ESP_LOGE("ESP32_NVS","nvs_get_i32 fail: %s %s", key, esp_err_to_name(err));
    }
    return value;
}
size_t preferences_getBytesLength(const char* key){
    size_t len = 0;
    if( !key){
        return 0;
    }
    esp_err_t err = nvs_get_blob(my_handle, key, NULL, &len);
    if(err){
        ESP_LOGE("ESP32_NVS","nvs_get_blob len fail: %s %s", key, esp_err_to_name(err));
        return 0;
    }
    return len;
}
size_t preferences_getBytes(const char* key, void * buf, size_t maxLen)
{
  size_t len = preferences_getBytesLength(key);
    if(!len || !buf || !maxLen){
        return len;
    }
    if(len > maxLen){
       ESP_LOGE("ESP32_NVS","not enough space in buffer: %u < %u", maxLen, len);
        return 0;
    }
    esp_err_t err = nvs_get_blob(my_handle, key, buf, &len);
    if(err){
        ESP_LOGE("ESP32_NVS","nvs_get_blob fail: %s %s", key, esp_err_to_name(err));
        return 0;
    }
    return len;
}

int32_t preferences_getFloat(const char* key, const float_t defaultValue)
{
   float_t value = defaultValue;
    preferences_getBytes(key, (void*) &value, sizeof(float_t));
    return value;
}
void calibrationGet(int16_t *gyro_x_offset, int16_t *gyro_y_offset,
                      int16_t *gyro_z_offset, float *angle_center)
{
  *gyro_x_offset = preferences_getInt("gryo_x", 0);
  *gyro_y_offset = preferences_getInt("gryo_y", 0);
  *gyro_z_offset = preferences_getInt("gryo_z", 0);

  *angle_center = preferences_getFloat("angle", 0.0);
}

/* End of NVS data reading functions */
void Imuinit(){
	int16_t x_offset1;
   	int16_t y_offset1;
   	int16_t z_offset1;
    float angle_offset1;

    calibrationInit();
    calibrationGet(&x_offset1,&y_offset1,&z_offset1,&angle_offset1);
    printf("IMU Calibrattion:\n");
    printf("x offset:%d, y offset:%d, z offset:%d, angle offset:%f\n",x_offset1,y_offset1,z_offset1,angle_offset1);
    i2c_mutex = xSemaphoreCreateMutex();
    SetMutex(&i2c_mutex);
    ImuTaskStart(x_offset1, y_offset1, z_offset1, &i2c_mutex);
}
void ImuTaskStart(int16_t x_offset, int16_t y_offset, int16_t z_offset, SemaphoreHandle_t *i2c_lock) {

    gryo_x_offset = 0 - x_offset;
    gryo_y_offset = 0 - y_offset;
    gryo_z_offset = 0 - z_offset;
    xTaskCreatePinnedToCore(ImuUpdateTask, "imu_task", 4 * 1024, i2c_lock, 5,
                            NULL, 1);
    
}

void I2C_Read_NBytes(uint8_t driver_Addr, uint8_t start_Addr,
                              uint8_t number_Bytes, uint8_t* read_Buffer) {
     uint8_t write_buffer[1];
    uint8_t read_buffer1[number_Bytes];
    uint8_t i2c_readI2CMaster_status;
    write_buffer[0] = start_Addr;            /* Wr0 */
    i2c_readI2CMaster_status = i2c_read_write(driver_Addr,I2C_NUM_0,write_buffer,
      sizeof(write_buffer), read_buffer1, sizeof(read_buffer1) );
//    i2c_ccode_B.I2CMaster_o1 = i2c_readI2CMaster_status;
    if (i2c_readI2CMaster_status == 0) {
    //   i2c_ccode_B.I2CMaster_o2 = read_buffer[0];/* Rd0 */
    //   i2c_ccode_B.I2CMaster_o3 = read_buffer[1];/* Rd1 */
    //   i2c_ccode_B.I2CMaster_o4 = read_buffer[2];/* Rd2 */
    //   i2c_ccode_B.I2CMaster_o5 = read_buffer[3];/* Rd3 */
    //   i2c_ccode_B.I2CMaster_o6 = read_buffer[4];/* Rd4 */
    //   i2c_ccode_B.I2CMaster_o7 = read_buffer[5];/* Rd5 */
      for (uint8_t i = 0; i < number_Bytes; i++) {
        read_Buffer[i] = read_buffer1[i];  // Copy data from read_buffer1 to read_Buffer
    }
    }
}

void I2C_Write_NBytes(uint8_t driver_Addr, uint8_t start_Addr,
                               uint8_t number_Bytes, uint8_t* write_Buffer) {
    uint8_t write_buffer1[number_Bytes + 1];
    uint8_t read_buffer[0];
    //uint8_t i2c_writeI2CMaster_status;
    write_buffer1[0] = start_Addr;           /* Wr0 */
    for (uint8_t i = 1; i <= number_Bytes; i++) {
        write_buffer1[i] = write_Buffer[i - 1];  // Copy data from write_Buffer to write_buffer1
    }
    i2c_read_write(driver_Addr,I2C_NUM_0,write_buffer1,
    sizeof(write_buffer1), read_buffer, sizeof(read_buffer) );
    //i2c_ccode_B.I2CMaster = i2c_writeI2CMaster_status;
}
void setFIFOEnable(_Bool enableflag) {
    uint8_t regdata = 0;
    regdata         = enableflag ? 0x18 : 0x00;
    I2C_Write_NBytes(MPU6886_ADDRESS, MPU6886_FIFO_ENABLE, 1, &regdata);
    regdata = enableflag ? 0x40 : 0x00;
    I2C_Write_NBytes(MPU6886_ADDRESS, MPU6886_USER_CTRL, 1, &regdata);
}
void ReadFIFOBuff(uint8_t* DataBuff, uint16_t Length) {
    uint8_t number = Length / 210;
    for (uint8_t i = 0; i < number; i++) {
        I2C_Read_NBytes(MPU6886_ADDRESS, MPU6886_FIFO_R_W, 210,
                        &DataBuff[i * 210]);
    }

    I2C_Read_NBytes(MPU6886_ADDRESS, MPU6886_FIFO_R_W, Length % 210,
                    &DataBuff[number * 210]);
}

void RestFIFO() {
    uint8_t buf_out;
    I2C_Read_NBytes(MPU6886_ADDRESS, MPU6886_USER_CTRL, 1, &buf_out);
    buf_out |= 0x04;
    I2C_Write_NBytes(MPU6886_ADDRESS, MPU6886_USER_CTRL, 1, &buf_out);
}

uint16_t ReadFIFOCount() {
    uint8_t Buff[2];
    uint16_t ReData = 0;
    I2C_Read_NBytes(MPU6886_ADDRESS, MPU6886_FIFO_COUNT, 2, Buff);
    ReData = (Buff[0] << 8) | Buff[1];
    return ReData;
}

void ImuUpdateTask(void *arg) {
    uint32_t last_ticks = xTaskGetTickCount();
    //printf("imuupdatetask\n");
    uint8_t *fifo_buff   = NULL;
    uint16_t fifo_count  = 0;
    uint16_t data_number = 0;
    fifo_buff            = (uint8_t *)malloc(sizeof(uint8_t) * 1024);

    float acc_x = 0, acc_y = 0, acc_z = 0;
    float gyro_x, gyro_y, gyro_z;
    float yaw_ahrs, pitch_ahrs, roll_ahrs;

    ImuData_t *imu_data;
    SemaphoreHandle_t i2c_lock = *(SemaphoreHandle_t *)arg;
    angle_lock                 = xSemaphoreCreateRecursiveMutex();

    bool fast_to_normal_angle = true;

    Butter_t butter_acc_x;
    Butter_t butter_acc_y;
    Butter_t butter_acc_z;

    memset(&butter_acc_x, 0, sizeof(Butter_t));
    memset(&butter_acc_y, 0, sizeof(Butter_t));
    memset(&butter_acc_z, 0, sizeof(Butter_t));

    // 500hz in, filter 50hz, low pass, butterworth
    butter_acc_x.a1 = -1.142980f;
    butter_acc_x.a2 = 0.412801f;
    butter_acc_x.b0 = 0.067455f;
    butter_acc_x.b1 = 0.134910f;
    butter_acc_x.b2 = 0.067455f;

    memcpy(&butter_acc_y, &butter_acc_x, sizeof(Butter_t));
    memcpy(&butter_acc_z, &butter_acc_x, sizeof(Butter_t));

    // Read from fifo to ensure that data will not be lost
    // IMU data freq is 500HZ
    xSemaphoreTake(i2c_lock, portMAX_DELAY);
    setFIFOEnable(true);
    RestFIFO();
    xSemaphoreGive(i2c_lock);

    // make sure angle fast to normal
    MadgwickAHRSetBeta(10);

    for (;;) {
        xSemaphoreTake(i2c_lock, portMAX_DELAY);
        fifo_count = ReadFIFOCount();
        ReadFIFOBuff(fifo_buff, fifo_count);
        xSemaphoreGive(i2c_lock);
        //printf("fifo_count: %d \n",fifo_count);

        imu_data    = (ImuData_t *)fifo_buff;
        data_number = fifo_count / 14;

        for (uint8_t i = 0; i < data_number; i++) {
            acc_x  = ACC_8G_RES * (int16_t)(imu_data[i].value.acc_x_l |
                                           (imu_data[i].value.acc_x_h << 8));
            acc_y  = ACC_8G_RES * (int16_t)(imu_data[i].value.acc_y_l |
                                           (imu_data[i].value.acc_y_h << 8));
            acc_z  = ACC_8G_RES * (int16_t)(imu_data[i].value.acc_z_l |
                                           (imu_data[i].value.acc_z_h << 8));
            gyro_x = GYRO_2000DPS_RES *
                     ((int16_t)(imu_data[i].value.gyro_x_l |
                                (imu_data[i].value.gyro_x_h << 8)) +
                      gryo_x_offset);
            gyro_y = GYRO_2000DPS_RES *
                     ((int16_t)(imu_data[i].value.gyro_y_l |
                                (imu_data[i].value.gyro_y_h << 8)) +
                      gryo_y_offset);
            gyro_z = GYRO_2000DPS_RES *
                     ((int16_t)(imu_data[i].value.gyro_z_l |
                                (imu_data[i].value.gyro_z_h << 8)) +
                      gryo_z_offset);

            // Don't know if this is necessary
            acc_x_2 = ButterFilter2(&butter_acc_x, acc_x);
            acc_y_2 = ButterFilter2(&butter_acc_y, acc_y);
            acc_z_2 = ButterFilter2(&butter_acc_z, acc_z);
            MadgwickAHRSupdateIMU(gyro_x * DEG_TO_RAD, gyro_y * DEG_TO_RAD,
                                  gyro_z * DEG_TO_RAD, acc_x_2, acc_y_2,
                                  acc_z_2, &pitch_ahrs, &roll_ahrs, &yaw_ahrs);
	        

        }

        if (fast_to_normal_angle) {
            int16_t roll = atan2(acc_y, sqrt(acc_z * acc_z + acc_x * acc_x)) *
                           360.0 / 2.0 / M_PI;

            if (fabs(roll - roll_ahrs) < 1 && roll != roll_ahrs) {
                fast_to_normal_angle = false;
                MadgwickAHRSetBeta(0.1);
            }
        }

        xSemaphoreTake(angle_lock, portMAX_DELAY);
        if (!fast_to_normal_angle) {
            angle = roll_ahrs;
        }
        xSemaphoreGive(angle_lock);
        vTaskDelayUntil(&last_ticks, pdMS_TO_TICKS(5));
    }
}
//====================================================================================================
// END OF CODE
//====================================================================================================